from __future__ import division # Makes it so 1/4 = 0.25 instead of zero
import numpy as np


from phidl import Device, Layer, quickplot
import phidl.geometry as pg
#==============================================================================
# We'll start by assuming we have a function waveguide() which already exists
# and makes us a simple waveguide rectangle
#==============================================================================

def waveguide(width = 10, height = 1):
    wg = Device('waveguide')
    wg.add_polygon( [(0, 0), (width, 0), (width, height), (0, height)] )
    wg.add_port(name = 'wgport1', midpoint = [0,height/2], width = height, orientation = 180)
    wg.add_port(name = 'wgport2', midpoint = [width,height/2], width = height, orientation = 0)
    return wg
   

#==============================================================================
# Create a blank device
#==============================================================================
# Create a new device ``D`` which will act as a blank canvas,
# and add a few waveguides to it to start out.  add_ref() returns
# the referenced object you added, allowing you to manipulate it later
D = Device('MultiWaveguide')

# We can instantiate the waveguide device by itself.  Note that when we make
# a device, we usually assign it a variable name with a capital letter
Wg1 = waveguide(width=10, height = 1)
Wg2 = waveguide(width=12, height = 2)

# 

wg1 = D.add_ref(Wg1)
wg2 = D.add_ref(Wg2)
wg3 = D.add_ref(waveguide(width=14, height = 3))

quickplot(D)


#==============================================================================
# Creating polygons
#==============================================================================
# Create and add a polygon from separate lists of x points and y points
# e.g. [(x1, x2, x3, ...), (y1, y2, y3, ...)]
poly1 = D.add_polygon( [(8,6,7,9), (6,8,9,5)] )

# Alternatively, create and add a polygon from a list of points
# e.g. [(x1,y1), (x2,y2), (x3,y3), ...] using the same function
poly2 = D.add_polygon( [(0, 0), (1, 1), (1, 3), (-3, 3)] )

quickplot(D)


#==============================================================================
# Manipulating geometry 1 - Basic movement and rotation
#==============================================================================
# There are several actions we can take to move and rotate the geometry.  These
# actions include movement, rotation, and reflection.

wg1.move([10,4]) # Shift the second waveguide we created over by dx = 10, dy = 4
wg2.move(origin = [1,1], destination = [2,2]) # Shift the second waveguide over by dx = 1, dy = 1
wg3.move([1,1], [5,5], axis = 'y') # Shift the third waveguide over by dx = 0, dy = 4 (motion only along y-axis)
poly1.movey(4) # Same as specifying axis='y' in the move() command
poly2.movex(4) # Same as specifying axis='x'' in the move() command
wg3.movex(30,40) # Moves "from" x=30 "to" x=40 (e.g. shifts wg3 by +10 in the x-direction)

wg1.rotate(45) # Rotate the first waveguide by 45 degrees around (0,0)
wg2.rotate(30, center = [1,1]) # Rotate the second waveguide by 30 degrees around (1,1)

wg1.reflect(p1 = [1,1], p2 = [1,3]) # Reflects wg3 across the line formed by p1 and p2


#==============================================================================
# Manipulating geometry 2 - Properties
#==============================================================================
# Each Device and DeviceReference object has several properties which can be used to learn
# information about the object (for instance where it's center coordinate is).  Several
# of these properties can actually be used to move the geometry by assigning them
# new values

wg1.bbox # Will return the bounding box of wg1 in terms of [(xmin, ymin), (xmax, ymax)]
wg1.xsize # Will return the width of wg1 in the x dimension
wg1.ysize # Will return the height of wg1 in the y dimension

wg1.center # Gives you the center coordinate of its bounding box
wg1.center = [4,4] # Shift wg1 such that the center coordinate of its bounding box is at (4,4)

wg2.xmax # Gives you the rightmost (+x) edge of the wg2 bounding box
wg2.xmax = 25 # Moves wg2 such that it's rightmost edge is at x = 25

wg3.ymin # Gives you the bottommost (+y) edge of the wg3 bounding box
wg3.ymin = -14 # Moves wg3 such that it's bottommost edge is at y = -14


quickplot(D)


#==============================================================================
# Manipulating geometry 3 - Smarter movement with ports
#==============================================================================
# All the waveguides we made have two ports: 'wgport1' and 'wgport2'  We can 
# use these names in place of (x,y) pairs.  For instance, if we want to move
# wg1 such that its port 'wgport1' rests on the origin, we do:
wg1.move(origin = 'wgport1', destination = [0,0])
# Alternatively, we can use the Port object itself in the same manner.  We can
# access the Port objects for any Device (or DeviceReference) by calling device.ports,
# --which returns a Python dictionary--and accessing its value with the key
wg3.move(origin = wg3.ports['wgport1'], destination = [0,0])
# We can even move one port to another 
wg2.move(origin = wg2.ports['wgport1'], destination = wg3.ports['wgport2'])
# Several functions beyond just move() can take Ports as inputs
wg1.rotate(angle = -60, center = wg1.ports['wgport2'])
wg3.reflect(p1 = wg3.ports['wgport1'].midpoint, p2 = wg3.ports['wgport1'].midpoint + np.array([1,0]))

quickplot(D)


#==============================================================================
# Manipulating geometry 4 - Chaining commands
#==============================================================================
# Many of the functions in Device return the object they manipulate.  We can use
# this to chain commands in a single line. For instance this:
wg1.rotate(angle = 15, center = [0,0])
wg1.move([10,20])
# ... is equivalent to this expression
wg1.rotate(angle = 15, center = [0,0]).move([10,20])



#==============================================================================
# Connecting devices with connect()
#==============================================================================
wg1.connect(port = 'wgport1', destination = wg2.ports['wgport2'])
wg3.connect(port = 'wgport2', destination = wg2.ports['wgport1'])

quickplot(D)



#==============================================================================
# Adding ports
#==============================================================================
# Although our waveguides have ports, ``D`` itself does not -- it only draws
# the subports (ports of wg1, wg2, wg3) as a convience.  We need to add ports
# that we specifically want in our new device ``D``
D.add_port(port = wg1.ports['wgport2'], name = 1)
D.add_port(port = wg3.ports['wgport1'], name = 2)

quickplot(D)



#==============================================================================
# Taking things a level higher
#==============================================================================
# Now that we have our device ``D`` which is a multi-waveguide device, we
# can add references to that device in a new blank canvas we'll call ``D2``.
# We'll add two copies of ``D`` to D2, and shift one so we can see them both
D2 = Device('MultiMultiWaveguide')
mwg1 = D2.add_ref(D)
mwg2 = D2.add_ref(D)
mwg2.move(destination = [10,10])

quickplot(D2)

# Like before, let's connect mwg1 and mwg2 together then offset them slightly
mwg1.connect(port = 1, destination = mwg2.ports[2])
mwg2.move(destination = [30,30])

quickplot(D2)


#==============================================================================
# Routing
#==============================================================================
# Routing allows us to connect two ports which face each other with a smooth
# polygon.  Since we connected our two 
D2.add_ref( pg.route(port1 = mwg1.ports[1], port2 = mwg2.ports[2], path_type = 'sine', width_type = 'straight') )
quickplot(D2)


#==============================================================================
# Adding text
#==============================================================================
# The function text() creates a Device, just like waveguide.  Use it and 
# manipulate it like any other Device
t = D2.add_ref( pg.text('Hello\nworld!', size = 10, justify = 'center'))
t.move([0,40]).rotate(45)
quickplot(D2)



#==============================================================================
# Labeling
#==============================================================================
# This label will display in a GDS viewer, but will not be rendered
# or printed like the polygons created by the text()
D2.annotate('First label', mwg1.center)
D2.annotate('Second label', mwg2.center)


#==============================================================================
# Saving the file as a .gds
#==============================================================================
D2.write_gds('MultiMultiWaveguideTutorial.gds')



#==============================================================================
# Using Layers
#==============================================================================
# Let's make a new blank device DL and add some text to it, but this time on
# different layers
DL = Device()

########
# You can specify any layer in one of three ways:
########
# 1) as a single number 0-255 representing the gds layer number, e.g. layer = 1
#    where the gds layer datatype will be automatically set to zero
DL.add_ref( pg.text('Layer1', size = 10, layer = 1) )


# 2) as a 2-element list [0,1] or tuple (0,1) representing the gds layer 
#    number (0-255) and gds layer datatype (0-255)  
DL.add_ref( pg.text('Layer2', size = 10, layer = [2,5]) ).movey(-20)


# 3) as a Layer object  
gold = Layer(name = 'goldpads', gds_layer = 3, gds_datatype = 0,
                 description = 'Gold pads liftoff', inverted = False)
DL.add_ref( pg.text('Layer3', size = 10, layer = gold) ).movey(-40)


# What you can also do is make a dictionary of layers, which lets you
# conveniently call each Layer object just by its name
layers = {
        'titanium' : Layer(gds_layer = 4, gds_datatype = 1, description = 'Gold pads liftoff', inverted = False),
        'niobium'  : Layer(gds_layer = 5, gds_datatype = 2, description = 'Gold pads liftoff', inverted = False),
        'nb_etch'  : Layer(gds_layer = 6, gds_datatype = 3, description = 'Niobium etch', inverted = False),
         }

# Now that our layers are defined, we can pass them to our text function
DL.add_ref( pg.text('Titanium', size = 10, layer = layers['titanium']) ).movey(-60)
DL.add_ref( pg.text('Niobium', size = 10, layer = layers['niobium']) ).movey(-80)
DL.add_ref( pg.text('Nb Etch', size = 10, layer = layers['nb_etch']) ).movey(-100)

quickplot(DL)

DL.write_gds('MultipleLayerText.gds')



from __future__ import unicode_literals, division, print_function, absolute_import

import io
import glob
import os
import pypandoc
try:
    pypandoc.get_pandoc_version()
except:
    print('pandoc installation not found, attempting to download it')
    pypandoc.download_pandoc()
    pypandoc.get_pandoc_version()

    
    
class Layer(object):
    def __init__(self, layer, datatype, shortname, description):
        self.gds_layer = gds_layer
        self.shortname = shortname
        self.description = description
    
    
class ProcessStep(object):
    
    def __init__(self, recipe, description, gds_layer):
        self.recipe = recipe
        self.description = description
        self.gds_layer = gds_layer
        
        
class Process(object):
    
    def __init__(self, recipes_dir, gds_files = None):
        self.steps = []
        self.gds_files = gds_files
        self.recipes_dir = recipes_dir
        self.recipes = self.load_recipes_directory(recipes_dir)

    def add_step(self, recipe, description, gds_layer):
        if recipe not in self.recipes:
            raise ValueError('No step named "%s" exists in the loaded recipes. Please check %s' % (recipe,self.recipes_dir))
        new_step = ProcessStep(recipe, description, gds_layer)
        self.steps.append(new_step)

        
    def check_gds_layers(self):
        layers_used = []
        for s in self.steps:
            l = s.gds_layer
            if l in layers_used:
                print('Warning: Layer %s is being used multiple times (in recipe %s)' % (l, s.recipe))
            layers_used.append(l)
            
            
    def write_pdf(self, filename = None):
        footer = ''
        md_output = ''
        for n, s in enumerate(self.steps):
            header = """\n\\newpage\n\n# Processing step %s (%s)\n\n-----\n\n""" % (n+1, s.recipe)
            instructions = self.recipes[s.recipe]
            md_output += header + instructions + footer
            
        pypandoc.convert_text(source = md_output, format = 'markdown', to = 'pdf', 
                              outputfile='hello214214.pdf', extra_args=['-s'])
        return md_output
        
    
#    def load_recipes_multiple_single_file(self, filename):
#        with open(filename, 'r') as text_file:
#            sections = text_file.read().split('[//]: # ')[1:]
#
#        for s in sections:
#            first_line_index = s.index('\n')
#            name = s[:first_line_index].strip(')( ')
#            instructions = s[first_line_index:].strip()
#            if name in self.recipes: raise ValueError('Duplicate recipe name found ()')
#            self.recipes[name] = instructions
            
    def load_recipes_directory(self, directory):
        recipes = {}
        recipe_files = glob.glob(directory + '*.md')
        for filepath in recipe_files:
            name = os.path.basename(filepath)
            name = os.path.splitext(name)[0]
            with open(filepath, 'r') as f:
                instructions = f.read()
            recipes[name] = instructions
        return recipes



gds_files = [
'SE0004.gds',
'SE0005.gds',
'SE0006.gds',
'SE0007.gds',
        ]
          
P = Process(recipes_dir = 'C:/Users/anm16/amcc/Code/device-layout/phidl/recipes/', gds_files = gds_files)

# Create patterned gold and dielectric layer
P.add_step(recipe = 'stepper_setup_01', description = 'Stepper setup', gds_layer = None)
P.add_step(recipe = 'gold_liftoff_01', description = 'Gold ground', gds_layer = 3)
P.check_gds_layers()
md = P.write_pdf(1)

#P.add_step(recipe = 'sio2_01', description = 'SiO2 dielectric', gds_layer = None)
#
## Do
#P.add_step(name = 'wsi_deposition_01', description = 'WSi deposition for SNSPDs', gds_layer = None)
#P.add_step(name = 'wsi_nanowires_01', description = 'WSi nanowire patterning', gds_layer = 0)
#P.add_step(name = 'wsi_etch_01', description = 'WSi SNSPD', gds_layer = 0)
#P.add_step(name = 'gold_liftoff_01', description = 'Gold contacts', gds_layer = 3, gds_inverted = True)

#%% 

import io
import pypandoc
try:
    pypandoc.get_pandoc_version()
except:
    print('pandoc installation not found, attempting to download it')
    pypandoc.download_pandoc()
    pypandoc.get_pandoc_version()


filename = 'C:\\Users\\anm16\\amcc\\Code\\device-layout\\phidl\\testprocess\\processes.md'
with io.open(filename, 'r') as text_file:
    md = text_file.read()
pypandoc.convert_text(source = md, format = 'markdown', to = 'pdf', 
                      outputfile='hello214214.pdf', extra_args=['-s'])

#%% Testing pdfkit

#import pdfkit
#pdfkit.from_string(html, 'out32153.pdf')


#%% Testing load functionality


filename = r'C:\Users\anm16\amcc\Code\device-layout\phidl\testprocess\processes.md'
with open(filename, 'r') as text_file:
    sections = text_file.read().split('[//]: # ')[1:]

names = []
texts = []
for s in sections:
    first_line_index = s.index('\n')
    name = s[:first_line_index].strip(')( ')
    text = s[first_line_index:].strip()
    names.append(name)
    texts.append(text)
    
#%% Trying YAML


import yaml

filename = 'C:\\Users\\anm16\\amcc\\Code\\device-layout\\phidl\\testprocess\\processes.yaml'
with open(filename) as f:
    y = yaml.load(f)

pypandoc.convert_text(source = y['wsi_deposition_01']['instructions'], format = 'markdown', to = 'pdf', 
                      outputfile='hello63414214.pdf', extra_args=['-s'])



#==============================================================================
# Potential improvements
#==============================================================================

# Add __or__, __and__, etc functionality to allow boolean operations
# Create DeviceSet
# Replace origin/destination with source/destination
# 

#==============================================================================
# Major TODO
#==============================================================================

 # TODO Add flatten()
 # TODO Move route to geometry

#==============================================================================
# Imports
#==============================================================================

from __future__ import division # Otherwise integer division e.g.  20 / 7 = 2
from __future__ import print_function # Use print('hello') instead of print 'hello'
from __future__ import absolute_import

import gdspy
import yaml
from copy import deepcopy
import numpy as np
from numpy import sqrt, mod, pi, sin, cos
from numpy.linalg import norm

from matplotlib import pyplot as plt
from matplotlib.patches import Polygon as PolygonPatch
from matplotlib.collections import PatchCollection

__version__ = '0.5.2'

#==============================================================================
# Useful transformation functions
#==============================================================================

def rotate_points(points, angle = 45, center = (0,0)):
    """ Rotates points around a centerpoint defined by ``center``.  ``points`` may be
    input as either single points [1,2] or array-like[N][2], and will return in kind
    """
    angle = angle*pi/180
    ca = cos(angle)
    sa = sin(angle)
    sa = np.array((-sa, sa))
    c0 = np.array(center)
    if np.array(points).ndim == 2: 
        return (points - c0) * ca + (points - c0)[:,::-1] * sa + c0
    if np.array(points).ndim == 1: 
        return (points - c0) * ca + (points - c0)[::-1] * sa + c0
    
def reflect_points(points, p1 = (0,0), p2 = (1,0)):
    """ Reflects points across the line formed by p1 and p2.  ``points`` may be
    input as either single points [1,2] or array-like[N][2], and will return in kind
    """
    # From http://math.stackexchange.com/questions/11515/point-reflection-across-a-line
    points = np.array(points); p1 = np.array(p1); p2 = np.array(p2);
    if np.array(points).ndim == 1: 
        return 2*(p1 + (p2-p1)*np.dot((p2-p1),(points-p1))/norm(p2-p1)**2) - points
    if np.array(points).ndim == 2: 
        return np.array([2*(p1 + (p2-p1)*np.dot((p2-p1),(p-p1))/norm(p2-p1)**2) - p for p in points])

def translate_points(points, d = [1,2]):
    """ Reflects points across the line formed by p1 and p2.  ``points`` may be
    input as either single points [1,2] or array-like[N][2], and will return in kind
    """
    points = np.array(points) + d
    return points
    

class Layer(object):
    def __init__(self, name = 'goldpads', gds_layer = 0, gds_datatype = 0,
                 description = 'Gold pads liftoff', inverted = False,
                 color = (.5, .5, .5)):
        self.name = name
        self.gds_layer = gds_layer
        self.gds_datatype = gds_datatype
        self.description = description
        
    
class _GeometryHelper(object):
    """ This is a helper class. It can be added to any other class which has 
    the functions move() and the property ``bbox`` (as in self.bbox).  It uses
    that function+property to enable you to do things like check what the center
    of the bounding box is (self.center), and also to do things like move the
    bounding box such that its maximum x value is 5.2 (self.xmax = 5.2) """
    
    @property
    def center(self):
        return np.sum(self.bbox,0)/2

    @center.setter
    def center(self, destination):
        self.move(destination = destination, origin = self.center)
        
    @property
    def x(self):
        return np.sum(self.bbox,0)[0]/2

    @x.setter
    def x(self, destination):
        destination = (destination, self.center[1])
        self.move(destination = destination, origin = self.center, axis = 'x')
        
    @property
    def y(self):
        return np.sum(self.bbox,0)[1]/2

    @y.setter
    def y(self, destination):
        destination = ( self.center[0], destination)
        self.move(destination = destination, origin = self.center, axis = 'y')
        
    @property
    def xmax(self):
        return self.bbox[1][0]

    @xmax.setter
    def xmax(self, destination):
        self.move(destination = (destination, 0), origin = self.bbox[1], axis = 'x')
        
    @property
    def ymax(self):
        return self.bbox[1][1]

    @ymax.setter
    def ymax(self, destination):
        self.move(destination = (0, destination), origin = self.bbox[1], axis = 'y')
        
    @property
    def xmin(self):
        return self.bbox[0][0]

    @xmin.setter
    def xmin(self, destination):
        self.move(destination = (destination, 0), origin = self.bbox[0], axis = 'x')
        
    @property
    def ymin(self):
        return self.bbox[0][1]

    @ymin.setter
    def ymin(self, destination):
        self.move(destination = (0, destination), origin = self.bbox[0], axis = 'y')
        
    @property
    def xsize(self):
        bbox = self.bbox
        return bbox[1][0] - bbox[0][0]
        
    @property
    def ysize(self):
        bbox = self.bbox
        return bbox[1][1] - bbox[0][1]

    def movex(self, origin = 0, destination = None):
        if destination is None:
            destination = origin
            origin = 0
        self.move(origin = (origin,0), destination = (destination,0))
        return self

    def movey(self, origin = 0, destination = None):
        if destination is None:
            destination = origin
            origin = 0
        self.move(origin = (0,origin), destination = (0,destination))
        return self
        




class Port(object):
    def __init__(self, name = None, midpoint = (0,0), width = 1, orientation = 90, parent = None):
        self.name = name
        self.midpoint = np.array(midpoint, dtype = 'float64')
        self.width = width
        self.orientation = mod(orientation,360)
        self.parent = parent
        if self.width <= 0: raise ValueError('[DEVICE] Port creation error: width cannot be negative or zero')
        
    def __repr__(self):
        return ('Port (name %s, midpoint %s, width %s, orientation %s)' % \
                (self.name, self.midpoint, self.width, self.orientation))
       
    @property
    def endpoints(self):
        dx = self.width/2*np.cos((self.orientation - 90)*pi/180)
        dy = self.width/2*np.sin((self.orientation - 90)*pi/180)
        left_point = self.midpoint - np.array([dx,dy])
        right_point = self.midpoint + np.array([dx,dy])
        return np.array([left_point, right_point])
    
    # FIXME currently broken
    @endpoints.setter
    def endpoints(self, points):
        p1, p2 = np.array(points[0]), np.array(points[1])
        self.midpoint = (p1+p2)/2
        dx, dy = p2-p1
        self.orientation = np.arctan2(-dy,dx)*180/pi
        self.width = sqrt(dx**2 + dy**2)
        
    @property
    def normal(self):
        dx = np.cos((self.orientation)*pi/180)
        dy = np.sin((self.orientation)*pi/180)
        return np.array([self.midpoint, self.midpoint + np.array([dx,dy])])
        
        


class Polygon(gdspy.Polygon, _GeometryHelper):
    @property
    def bbox(self):
        return np.asarray( (np.min(self.points, axis = 0), np.max(self.points, axis = 0)))

    def rotate(self, angle = 45, center = (0,0)):
        super(Polygon, self).rotate(angle = angle*pi/180, center = center)
        return self
            
    def move(self, origin = (0,0), destination = None, axis = None):
        """ Moves elements of the Device from the origin point to the destination.  Both
         origin and destination can be 1x2 array-like, Port, or a key
         corresponding to one of the Ports in this device """

        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = [0,0]

        if isinstance(origin, Port):            o = origin.midpoint
        elif np.array(origin).size == 2:    o = origin
        elif self.ports.has_key(origin):    o = self.ports[origin].midpoint
        else: raise ValueError('[DeviceReference.move()] ``origin`` not array-like, a port, or port name')
            
        if isinstance(destination, Port):           d = destination.midpoint
        elif np.array(destination).size == 2:        d = destination
        elif self.ports.has_key(destination):   d = self.ports[destination].midpoint
        else: raise ValueError('[DeviceReference.move()] ``destination`` not array-like, a port, or port name')

        if axis == 'x': d = (d[0], o[1])
        if axis == 'y': d = (o[0], d[1])

        dx,dy = np.array(d) - o

        super(Polygon, self).translate(dx, dy)
        return self

            
    def reflect(self, p1 = (0,1), p2 = (0,0)):
        self.points = reflect_points(self.points, p1, p2)
        return self


        
class Device(gdspy.Cell, _GeometryHelper):
    uid = 0
    
    def __init__(self, name = 'Unnamed'):
        self.ports = {}
        self.parameters = {}
        self.meta = {}
        self.references = []

        Device.uid += 1
        name = '%s%06d' % (name[:20], Device.uid) # Write name e.g. 'Unnamed000005'

        super(Device, self).__init__(name, exclude_from_global=True)

    @property
    def layers(self):
        return self.get_layers()

    @property
    def bbox(self):
        self.bb_is_valid = False # IMPROVEMENT This is a hack to get around gdspy caching issues
        return np.array(self.get_bounding_box())

        
    def add_ref(self, device, config = None, **kwargs):
        """ Takes a Device (or Device-making function with config) and adds it
        as a DeviceReference to the current Device.  """
         # Check if ``device`` is actually a device-making function
        if callable(device):    d = makedevice(fun = device, config = config, **kwargs)
        else:                   d = device
        device_ref = DeviceReference(d)   # Create a DeviceReference (CellReference)
        self.add(device_ref)             # Add DeviceReference (CellReference) to Device (Cell)
        self.references.append(device_ref) # Add to the list of references (for convenience)
        return device_ref                # Return the DeviceReference (CellReference)


    def add_polygon(self, points, layer = 0):
        if isinstance(points, gdspy.Polygon):
            points = points.points
        elif isinstance(points, gdspy.PolygonSet):
            return [self.add_polygon(p, layer) for p in points.polygons]
                
        # Check if the variable layer is a Layer object, a 2-element list like
        # [0,1] representing layer=0 and datatype=1, or just a layer number
        if isinstance(layer, Layer):
            gds_layer = layer.gds_layer
            gds_datatype = layer.gds_datatype
        elif np.size(layer) == 2:
            gds_layer = layer[0]
            gds_datatype = layer[1]
        else:
            gds_layer = layer
            gds_datatype = 0
            
        
        if len(points[0]) == 2: # Then it must be of the form [[1,2],[3,4],[5,6]]
            polygon = Polygon(points, gds_layer, gds_datatype)
        elif len(points[0]) > 2: # Then it must be of the form [[1,3,5],[2,4,6]]
            polygon = Polygon(xy2p(points), gds_layer, gds_datatype)
        self.add(polygon)
        return polygon
        
        
    def add_port(self, name = None, midpoint = (0,0), width = 1, orientation = 45, port = None):
        """ Can be called to copy an existing port like add_port(port = existing_port) or
        to create a new port add_port(myname, mymidpoint, mywidth, myorientation).
        Can also be called to copy an existing port with a new name like add_port(port = existing_port, name = new_name)"""
        if isinstance(port, Port): p = deepcopy(port)
        elif isinstance(name, Port):
            p = deepcopy(name)
            name = p.name
        else:                  p = Port(name, midpoint, width, orientation, parent = self)
        if name is not None: p.name = name
        if self.ports.has_key(p.name):
            raise ValueError('[DEVICE] add_port() error: Port name already exists in this device') 
        self.ports[p.name] = p
        return p
        
        
    def add_array(self, device, start = (0,0), spacing = (10,0), num_devices = 6, config = None, **kwargs):
         # Check if ``device`` is actually a device-making function
        if callable(device):    d = makedevice(fun = device, config = config, **kwargs)
        else:                   d = device
        references = []
        for n in range(num_devices):
            sd = self.add_ref(d)
            sd.move(destination = np.array(spacing)*n, origin = -np.array(start))
            references.append(sd)
        return references
        

    def annotate(self, text = 'hello', position = (0,0), layer = 89):
        if type(text) is not str: text = str(text)
        l = self.add(gdspy.Label(text = text, position = position, anchor = 'o', layer=layer))
        return l
        
#
#    def delete_port(self, name):
#        self.ports.pop(name, None)
#    
    
    def write_gds(self, filename, unit = 1e-6, precision = 1e-9):
        if filename[-4:] != '.gds':  filename += '.gds'
        tempname = self.name
        self.name = 'toplevel'
        gdspy.gds_print(filename, cells=[self], name='library', unit=unit, precision=precision)
        self.name = tempname


    def distribute(self, elements, direction = 'x', spacing = 100, separation = True):
        multiplier = 1
        if   direction[0] == '+':
            direction = direction[1:]
        elif direction[0] == '-':
            direction = direction[1:]
            multiplier = -1

        xy = np.array([0,0])
        for e in elements:
            e.center = xy
            if direction == 'x':
                xy = xy + (np.array([spacing, 0]) + np.array([e.xsize, 0])*(separation==True))*multiplier
            elif direction == 'y':
                xy = xy + (np.array([0, spacing]) + np.array([0, e.ysize])*(separation==True))*multiplier
            else:
                raise ValueError('[PHIDL] distribute() needs a direction of "x", "+y", "-x", etc')

        
    def flatten(self, depth = None):
        pass

    
    def rotate(self, angle = 45, center = (0,0)):
        for e in self.elements:
            if isinstance(e, Polygon):
                e.rotate(angle = angle, center = center)
            elif isinstance(e, (gdspy.Polygon, gdspy.PolygonSet)):
                e.rotate(angle = angle*pi/180, center = center)
            elif isinstance(e, DeviceReference):
                e.rotate(angle, center)
        for p in self.ports.values():
            p.midpoint = rotate_points(p.midpoint, angle, center)
            p.orientation = mod(p.orientation + angle, 360)
        return self
            
    def move(self, origin = (0,0), destination = None, axis = None):
        """ Moves elements of the Device from the origin point to the destination.  Both
         origin and destination can be 1x2 array-like, Port, or a key
         corresponding to one of the Ports in this device """

        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = [0,0]

        if isinstance(origin, Port):            o = origin.midpoint
        elif np.array(origin).size == 2:    o = origin
        elif self.ports.has_key(origin):    o = self.ports[origin].midpoint
        else: raise ValueError('[DeviceReference.move()] ``origin`` not array-like, a port, or port name')
            
        if isinstance(destination, Port):           d = destination.midpoint
        elif np.array(destination).size == 2:        d = destination
        elif self.ports.has_key(destination):   d = self.ports[destination].midpoint
        else: raise ValueError('[DeviceReference.move()] ``destination`` not array-like, a port, or port name')

        if axis == 'x': d = (d[0], o[1])
        if axis == 'y': d = (o[0], d[1])

        dx,dy = np.array(d) - o
        
        # Move geometries
        for e in self.elements:
            if isinstance(e, (gdspy.Polygon, gdspy.PolygonSet)): 
                e.translate(dx,dy)
            if isinstance(e, DeviceReference): 
                e.move(destination = d, origin = o)
        for p in self.ports.values():
            p.midpoint = np.array(p.midpoint) + np.array(d) - np.array(o)
        
        # Move labels
        for l in self.labels:
            l.translate(dx,dy)
        
        return self
            
    def reflect(self, p1 = (0,1), p2 = (0,0)):
        for e in self.elements:
            if isinstance(e, gdspy.Polygon):
                e.points = reflect_points(e.points, p1, p2)
            elif isinstance(e, gdspy.PolygonSet):
                for poly in e.polygons:
                    poly.points = reflect_points(poly.points, p1, p2)
            elif isinstance(e, DeviceReference):
                e.reflect(p1, p2)
        for p in self.ports.values():
            p.midpoint = reflect_points(p.midpoint, p1, p2)
            phi = np.arctan2(p2[1]-p1[1], p2[0]-p1[0])*180/pi
            p.orientation = 2*phi - p.orientation
                
    
    
class DeviceReference(gdspy.CellReference, _GeometryHelper):
    def __init__(self, device, origin=(0, 0), rotation=0, magnification=None, x_reflection=False):
        super(DeviceReference, self).__init__(device, origin, rotation, magnification, x_reflection)
        self.parent = device
        self._parent_ports = device.ports
        self._local_ports = deepcopy(device.ports)
    
    @property
    def ports(self):
        """ This property allows you to access my_device_reference.ports, and receive a copy
        of the ports dict which is correctly rotated and translated"""
        for key in self._parent_ports.keys():
            port = self._parent_ports[key] 
            new_midpoint, new_orientation = self._transform_port(port.midpoint, \
                port.orientation, self.origin, self.rotation, self.x_reflection)
            self._local_ports[key].midpoint = new_midpoint
            self._local_ports[key].orientation = mod(new_orientation,360)
            self._local_ports[key].parent = self
        return self._local_ports

    @property
    def meta(self):
        return self.parent.meta
        
    @property
    def bbox(self):
        return self.get_bounding_box()

        
    def _transform_port(self, point, orientation, origin=(0, 0), rotation=None, x_reflection=False):
        # Apply GDS-type transformations (x_ref)
        new_point = np.array(point)
        new_orientation = orientation
        
        if x_reflection:
            new_point[1] = -new_point[1]
            new_orientation = -orientation
        if rotation is not None:
            new_point = rotate_points(new_point, angle = rotation, center = [0, 0])
            new_orientation += rotation
        if origin is not None:
            new_point = new_point + np.array(origin)
        new_orientation = mod(new_orientation, 360)
            
        return new_point, new_orientation

        
    def move(self, origin = (0,0), destination = None, axis = None):
        """ Moves the DeviceReference from the origin point to the destination.  Both
         origin and destination can be 1x2 array-like, Port, or a key
         corresponding to one of the Ports in this device_ref """

        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = (0,0)

        if isinstance(origin, Port):            o = origin.midpoint
        elif np.array(origin).size == 2:    o = origin
        elif self.ports.has_key(origin):    o = self.ports[origin].midpoint
        else: raise ValueError('[DeviceReference.move()] ``origin`` not array-like, a port, or port name')
            
        if isinstance(destination, Port):           d = destination.midpoint
        elif np.array(destination).size == 2:   d = destination
        elif self.ports.has_key(destination):   d = self.ports[destination].midpoint
        else: raise ValueError('[DeviceReference.move()] ``destination`` not array-like, a port, or port name')
            
        # Lock one axis if necessary
        if axis == 'x': d = (d[0], o[1])
        if axis == 'y': d = (o[0], d[1])

        # This needs to be done in two steps otherwise floating point errors can accrue
        dxdy = np.array(d) - np.array(o)
        self.origin = np.array(self.origin) + dxdy
        return self

        
    def rotate(self, angle = 45, center = (0,0)):
        if type(center) is Port:  center = center.midpoint
        self.rotation += angle
        self.origin = rotate_points(self.origin, angle, center)
        return self
        
        
    def reflect(self, p1 = (0,1), p2 = (0,0)):
        if type(p1) is Port:  p1 = p1.midpoint
        if type(p2) is Port:  p2 = p2.midpoint
        p1 = np.array(p1);  p2 = np.array(p2)
        # Translate so reflection axis passes through origin
        self.origin = self.origin - p1
        
        # Rotate so reflection axis aligns with x-axis
        angle = np.arctan((p2[1]-p1[1])/(p2[0]-p1[0]))*180/pi
        self.origin = rotate_points(self.origin, angle = -angle, center = [0,0])
        self.rotation -= angle
        
        # Reflect across x-axis
        self.x_reflection = not self.x_reflection
        self.origin[1] = -self.origin[1]
        self.rotation = -self.rotation
        
        # Un-rotate and un-translate
        self.origin = rotate_points(self.origin, angle = angle, center = [0,0])
        self.rotation += angle
        self.origin = self.origin + p1
        return self
        
        
    def connect(self, port, destination):
        # ``port`` can either be a string with the name or an actual Port
        if self.ports.has_key(port):
            p = self.ports[port]
        elif type(port) is Port:
            p = port
        else:
            raise ValueError('[PHIDL] connect() did not receive a Port or valid port name' + \
                ' - received (%s), ports available are (%s)' % (port, self.ports.keys()))

        self.rotate(angle =  180 + destination.orientation - p.orientation, center = p.midpoint)
        self.move(origin = p, destination = destination)
        return self


#==============================================================================
# Handy geometry functions
#==============================================================================


def _load_gds(filename, cell_name, load_ports = True):
    # Format "Port(name = hello, width = 20, orientation = 180)"
    gdspy.Cell.cell_dict.clear()
    gdsii = gdspy.GdsImport(filename)
    gdsii.extract(cell_name)
    d = Device(cell_name)
    d.elements = gdspy.Cell.cell_dict[cell_name].elements
    for label in gdspy.Cell.cell_dict[cell_name].labels:
        t = label.text
        if t[0:5] == 'Port(' and t[-1] == ')':
            arguments = t[5:-1]
            arguments = arguments.replace(' ', '')
            args = {a.split('=')[0] : a.split('=')[1] for a in arguments.split(',')}
            if args['name'].isdigit():args['name'] = int(args['name'])
            d.add_port(name = args['name'], midpoint = label.position, width = float(args['width']), orientation = float(args['orientation']))
        else:
            d.labels.append(label)
    return d


#==============================================================================
# Helper functions
#==============================================================================

def p2xy(points):
    """ Takes in a list of [x,y] pairs and converts them to lists of x points 
    and y points.  So p2xy([[1,5],[2,6],[3,7]]) returns [[1,2,3],[5,6,7]]
    """
    p = np.array(points)
    x = p[:,0]
    y = p[:,1]
    return np.array([x,y])
    
def xy2p(*args):
    """ Takes in lists of x points and y points, e.g. [1,2,3],[5,6,7] and
    converts it to the point format e.g. [[1,5],[2,6],[3,7]].  Can either be
    called as xy2p(xpts, ypts) or as xy2p(xy) where xy = [xpts, ypts]
    """
    if len(args) == 1:      x,y = args[0][0], args[0][1] 
    elif len(args) == 2:    x,y = args[0],    args[1]
    points = np.array(zip(*[x,y]))
    return points
    
    
def quickplot(items, overlay_ports = True, overlay_subports = True, label_ports = True, new_window = True):
    """ Takes a list of devices/references/polygons or single one of those, and
    plots them.  Also has the option to overlay their ports """
    if new_window: fig, ax = plt.subplots(1)
    else:
        ax = plt.gca()  # Get current figure
        ax.cla()        # Clears the axes of all previous polygons
    
    # Iterate through each each Device/DeviceReference/Polygon
    patches = []
    if type(items) is not list:  items = [items]
    for item in items:
        if isinstance(item, (Device, DeviceReference)):
            polygons = item.get_polygons(by_spec=False, depth=None)
            for p in polygons:
                patches.append(PolygonPatch(p, closed=True, alpha = 0.4))
            for name, port in item.ports.items():
                _draw_port(port, arrow_scale = 2, shape = 'full', color = 'k')
                plt.text(port.midpoint[0], port.midpoint[1], name)
        if isinstance(item, Device) and overlay_subports is True:
            for sd in item.references:
                for name, port in sd.ports.items():
                    _draw_port(port, arrow_scale = 1, shape = 'right', color = 'r')
                    plt.text(port.midpoint[0], port.midpoint[1], name)
        if isinstance(item, gdspy.Polygon):
            patches.append(PolygonPatch(item.points, closed=True, alpha = 0.4))
        if isinstance(item, gdspy.PolygonSet):
            for p in item.polygons:
                patches.append(PolygonPatch(p, closed=True, alpha = 0.4))
    pc = PatchCollection(patches, alpha=0.4)
    # TODO: Change this to per-layer coloring    
    np.random.seed(0)
    colors = 100*np.random.rand(len(patches))
    pc.set_array(np.array(colors))
    ax.add_collection(pc)
    plt.axis('equal')
    ax.grid(True, which='both', alpha = 0.4)
    ax.axhline(y=0, color='k', alpha = 0.2, linewidth = 1)
    ax.axvline(x=0, color='k', alpha = 0.2, linewidth = 1)
    plt.draw()


def _draw_port(port, arrow_scale = 1, **kwargs):
    x = port.midpoint[0]
    y = port.midpoint[1]
    nv = port.normal
    n = (nv[1]-nv[0])*arrow_scale
    dx, dy = n[0], n[1]
    xbound, ybound = p2xy(port.endpoints)
    #plt.plot(x, y, 'rp', markersize = 12) # Draw port midpoint
    plt.plot(xbound, ybound, 'r', linewidth = 3) # Draw port edge
    plt.arrow(x, y, dx, dy,length_includes_head=True, width = 0.1*arrow_scale, head_width=0.3*arrow_scale, **kwargs)


def makedevice(fun, config = None, **kwargs):
    config_dict = {}
    if type(config) is str:
        with open(config) as f:  config_dict = yaml.load(f) # Load arguments from config file
    elif type(config) is dict:   config_dict = config
    config_dict.update(**kwargs)
    return fun(**config_dict)
#
#    
#def useconfig(filename = 'myconfig.yaml', **kwargs):
#    with open(filename) as f:  config_dict = yaml.load(f) # Load arguments from config file
#    config_dict.update(**kwargs)                          # Replace any additional arguments  
#    return config_dict
#
#    
#filename = 'C:/Users/anm16/Downloads/temp.yaml'
#d = _load_config_file(filename)
#
#
#y = makedevice(beamsplitter, filename, arm_length = 50)
#quickplot(y)
#
#y = beamsplitter(**useconfig(filename, arm_length = 50))
#quickplot(y)





#!/usr/bin/env python

from setuptools import setup

install_requires=[
   'gdspy>=1.0',
   'numpy',
   'matplotlib',
   'pyyaml',
   'scikit-image>=0.11',
]

setup(name='phidl',
      version='0.5.2',
      description='PHIDL',
      install_requires=install_requires,
      author='Adam McCaughan',
      author_email='adam.mccaughan@nist.gov',
      packages=['phidl'],
      py_modules=['phidl.geometry'],
      package_dir = {'phidl': 'phidl'},
     )